part of validation;

class Success<F, S> extends Validation<F, S> {

  final S _data;

  /**
   * Const constructor with const initialization.
   */
  const Success(S this._data);

  /**
   * An approximation of pattern matching on the Validation type.
   *
   * Always calls the supplied successful computation with the inner wrapped
   * value of this Success and returns the result.
   */
  dynamic cata(dynamic failure(F fail), dynamic successful(S success)) =>
    successful(_data);

  /**
   * An optional failure projection.
   *
   * Returns the const instance of None<T>.
   */
  Option<F> failure() => const None();

  /**
   * An optional success projection.
   *
   * Returns a Some<S> instance containing the inner wrapped value of
   * this Success.
   */
  Option<S> successful() => new Some(_data);

  /**
   * Attempts to unwrap the successful case with a fallback computation.
   *
   * Returns the inner wrapped value.
   */
  S getOrElse(S alternative()) => _data;

  /**
   * Attempts to unwrap the successful case with a fallback value.
   *
   * Returns the inner wrapped value.
   */
  S getOrDefault(S def) => _data;

  /**
   * Computes an alternative for the failure case.
   *
   * Returns itself immediately.
   */
  Validation<F, S> orElse(Validation<F,S> alternative()) => this;

  /**
   * Performs a map on this Validation type.
   *
   * Returns a new instance of Success containing the result of calling the
   * supplied mapper with the inner wrapped value.
   */
  Validation<F, dynamic> map(dynamic mapper(S n)) =>
    new Success(mapper(_data));

  /**
   * Applies another applicative Validation to this one.
   *
   * Returns the result of calling map on the supplied Validation with the
   * inner wrapped function of this Validation.
   */
  Validation<F, dynamic> ap(Validation<F, dynamic> other) => other.map(_data);

  /**
   * Performs a monadic expansion on this Validation type.
   *
   * Returns the Validation instance returned from calling the supplied
   * expander with the inner wrapped value.
   */
  Validation<F, dynamic> expand(Validation<F, dynamic> expander(S n)) =>
    expander(_data);

  /**
   * Performs a map on the failure case.
   *
   * Returns itself immediately.
   */
  Validation<dynamic, S> leftMap(dynamic mapper(F f)) => this;

}