part of validation;

abstract class Validation<F, S>
implements Functor<Validation>,
           Applicative<Validation>,
           Monad<Validation>
{

  /**
   * An approximation of pattern matching on the Validation type.
   *
   * If this Validation is an instance of Failure, then the result of
   * calling `failure(fail)` with the inner wrapped value of the failure
   * will be returned. Otherwise, the result of calling `successful(success)`
   * with the inner wrapped value of the Success will be called.
   */
  dynamic cata(dynamic failure(F fail), dynamic successful(S success));

  /**
   * Creates an optional projection of the failure case.
   *
   * If this validation is an instance of Failure then an Option<F> is returned
   * containing the inner wrapped value of the Failure instance. Otherwise, if
   * this is an instance of Success then None<F> is returned.
   */
  Option<F> failure();

  /**
   * Creates an optional projection of the success case.
   *
   * If this validation is an instance of Success then an Option<S> is returned
   * containing the inner wrapped value of the Success instance. Otherwise, if
   * this is an instance of Failure then a None<S> is returned.
   */
  Option<S> successful();

  /**
   * Attempts to unwrap the successful case with a fallback computation.
   *
   * If this validation is an instance of Success then the inner wrapped value
   * will be returned. Otherwise if the validation is an instance of Failure
   * it will return the result of calling the supplied alternative computation.
   */
  S getOrElse(S alternative());

  /**
   * Attempts to unwrap the successful case with a fallback value.
   *
   * If this validation is an instance of Success then the inner wrapped value
   * will be returned. Otherwise if the validation is an instance of Failure
   * it will return the supplied alternative value.
   */
  S getOrDefault(S def);

  /**
   * Computes an alternative for the failure case.
   *
   * If this validation is an instance of Success then it returns itself
   * immediately. Otherwise if this is an instance Failure then the result
   * of calling the supplied alternative computation will be returned.
   */
  Validation<F, S> orElse(Validation<F,S> alternative());

  /**
   * Performs a map on the failure case.
   *
   * If this validation is an instance of Success then it returns itself
   * immediately. Otherwise, if this is an instance of Failure then a
   * new Failure will be returned containing the result of the mapper
   * applied to the inner wrapped value.
   */
  Validation<dynamic, S> leftMap(dynamic mapper(F failure));

}